/*
 * Flash memory interface rev.5 driver for the Intel 
 *
 */
 
#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <asm/segment.h>
//#include <unistd.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>

/*****************************************************************************/
#include "nwflash.h"

//#define MINIKERNEL 1		//export flash write, erase routines for MiniKernel

#ifndef MINIKERNEL
static
#endif
void kick_open(void);
#ifndef MINIKERNEL
static
#endif
int get_flash_id(void);
#ifndef MINIKERNEL
static
#endif
int erase_block(int nBlock);		
#ifndef MINIKERNEL
static
#endif
int write_block(unsigned long p, const char* buf, int count);		
static int open_flash(struct inode* inodep, struct file* filep);
static void release_flash(struct inode* inodep, struct file* filep);
static int flash_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg);
static int flash_read(struct inode * inode, struct file * file, char * buf, int count);
static int flash_write(struct inode * inode, struct file * file, const char * buf, int count);
static int flash_lseek(struct inode * inode, struct file * file, off_t offset, int orig);

#define	KFLASH_BASE 0xE1800000	//kernel virtual flash address
#define KFLASH_SIZE 1024*1024	//1 Meg
#define KFLASH_SIZE4 4*1024*1024//4 Meg
#define KFLASH_ID 0x89A6	//Intel flash
#define KFLASH_ID4 0xB0D4	//Intel flash 4Meg
#define KROM_ACCESS 0xE1000068	//the footbridge reg
#define FB_TIMER1 0xE1000304	//the footbridge timer 1 (shaken 66Mhz)
#define FB_TIMER3 0xE1000344	//the footbridge timer 3 (7.1 Mhz)
#define TIMER1_15us 1000

static int flashdebug=0;	//if set - we will display progress msgs
static int gbWriteEnable = 0;
static int gbWriteBase64Enable = 0;
static int gbFlashSize = KFLASH_SIZE;

static struct file_operations flash_fops = {
	flash_lseek,		/* lseek              */
	flash_read,		/* read               */
	flash_write,		/* write              */
	NULL,			/* no special readdir            */
	NULL,			/* no special select             */
	flash_ioctl,
	NULL,			/* no special mmap               */
	open_flash,
	release_flash,
	NULL,			/* no special fsync              */
	NULL,			/* no special fasync             */
	NULL,			/* no special check_media_change */
	NULL			/* no special revaldate          */
};

static struct miscdevice flash_miscdev=
{
	FLASH_MINOR,
	"nwflash",
	&flash_fops
};

#ifndef MINIKERNEL
static
#endif
int get_flash_id()
{
volatile unsigned int	c1,c2;
int	port338;
unsigned long	flags;

	// try to get flash chip ID 
	save_flags(flags);
	cli();

	port338 = inb(0x338);
	kick_open();
	c2 = inb(0x80);
	*(unsigned char *)(KFLASH_BASE+0x8000) = 0x90;
	udelay(15);
	c1 = *(unsigned char *)KFLASH_BASE;
	c2 = inb(0x80);

	//on 4 Meg flash the second byte is actually at offset 2...
	if (c1 == 0xB0)
		c2 = *(unsigned char *)(KFLASH_BASE+2);
	else
		c2 = *(unsigned char *)(KFLASH_BASE+1);
		
	c2 += (c1<<8);
	// set it back to read mode 
	*(unsigned char *)(KFLASH_BASE+0x8000) = 0xFF;
	outb(port338, 0x338);
	restore_flags(flags);

	if (c2 == KFLASH_ID4)
	    gbFlashSize = KFLASH_SIZE4;
		
	return(c2);
}	
	
static int open_flash(struct inode* inodep, struct file* filep)
{
int	id;

	id=get_flash_id();
	if ((id != KFLASH_ID) && (id != KFLASH_ID4))
	{
		printk("Flash: incorrect ID 0x%04X.\n",id);
		return -ENXIO;
	}
	MOD_INC_USE_COUNT;

	return 0;
}


static void release_flash(struct inode* inodep, struct file* filep)
{
	MOD_DEC_USE_COUNT;
}


static int flash_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg)
{
//	printk("Flash_ioctl: cmd = 0x%X.\n",cmd);

	switch (cmd)
	{
		case CMD_WRITE_DISABLE:
		    	gbWriteBase64Enable = 0;
			gbWriteEnable = 0;
			break;		

		case CMD_WRITE_ENABLE:
			gbWriteEnable = 1;
			break;

		case CMD_WRITE_BASE64K_ENABLE:
		    	gbWriteBase64Enable = 1;
			break;

		default:
		    	gbWriteBase64Enable = 0;
			gbWriteEnable = 0;
			return -EINVAL;
	}

	return 0;
}



static int flash_read(struct inode * inode, struct file * file, char * buf, int count)
{
	unsigned long p = file->f_pos;
	int read;

	if (flashdebug)
	printk("Flash_dev: flash_read: offset=0x%X, buffer=0x%X, count=0x%X.\n",
		    (unsigned int)p, (unsigned int)buf, count);

	
	if (count < 0)
		return -EINVAL;

	if (count > gbFlashSize - p)
		count = gbFlashSize - p;

	if (verify_area(VERIFY_WRITE, buf, count) == -EFAULT)
		return -EFAULT;
		
	p += KFLASH_BASE;	//flash virtual address
			
	read = 0;

	memcpy_tofs(buf, (void *) p, count);
	read += count;
	file->f_pos += read;
	return read;
}

static int flash_write(struct inode * inode, struct file * file, const char * buf, int count)
{
    unsigned long p = file->f_pos;
    int written;
    int nBlock,temp,rc;
    int	port338;
    int	i,j;

	
	if (flashdebug)
	printk("Flash_dev: flash_write: offset=0x%X, buffer=0x%X, count=0x%X.\n",
		    (unsigned int)p, (unsigned int)buf, count);
	
	if (!gbWriteEnable)
		return -EINVAL;
			
	if (p<64*1024 && (!gbWriteBase64Enable))
		return -EINVAL;
			
	if (count < 0)
		return -EINVAL;
		
	//if write size to big - error!
	if (count > gbFlashSize - p)
		return -EINVAL;

		
	if (verify_area(VERIFY_READ, buf, count) == -EFAULT)
		return -EFAULT;
		
		
	written = 0;
		
	port338 = inb(0x338);		//save port settings

			
	nBlock = (int)p>>16;		//block # of 64K bytes
	temp = ((int)(p+count)>>16)-nBlock+1; //# of 64K blocks to erase and write

	//write ends at exactly 64k boundry?
	if (((int)(p+count) & 0xFFFF)== 0)
	{
	    temp -= 1;
	
	}
			
	if (flashdebug)
	printk("FlashWrite: writing %d block(s) starting at %d.\n",temp,nBlock);	
	
	for( ;temp;temp--, nBlock++)		
	{
		if (flashdebug)
		    printk("FlashWrite: erasing block %d.\n",nBlock);
		
		//first we have to erase the block(s), where we will write...
		i = 0;
		j = 0;
RetryBlock:
		do
		{
		rc = erase_block(nBlock);
		i++;
		} while (rc && i<10);

		if (rc)
		{
		    if (flashdebug)
			printk("FlashWrite: erase error %X. Aborting...\n",rc);	
		    outb(port338, 0x338);	//restore reg on exit
		    return(written);	// bytes written
		}
		
		if (flashdebug)
		printk("FlashWrite: writing offset %X, from buf %X, bytes left %X.\n",
		    (unsigned int)p,(unsigned int)buf,count-written);

		//write_block will limit write to space left in this block
		rc = write_block(p,buf,count-written);
		j++;
				
		if (!rc)	//if somehow write verify failed? Can't happen??
		{
		    if (j < 10)	//retry up to 10 times
			goto RetryBlock;
		    else
			rc = -1;	//else quit with error...
		}
		
		if (rc < 0) 
		{
		    if (flashdebug)
			printk("FlashWrite: write error %X. Aborting...\n",rc);	
		    outb(port338, 0x338);	//restore reg on exit
		    return(written);	// bytes written
		}
		
		p += rc;		
		buf += rc;
		written += rc;
		file->f_pos += rc;

		if (flashdebug)
		    printk("FlashWrite: written 0x%X bytes OK.\n",written);	
	}		

        outb(port338, 0x338);		//restore reg on exit
	return written;
}


/*
 * The memory devices use the full 32/64 bits of the offset, and so we cannot
 * check against negative addresses: they are ok. The return value is weird,
 * though, in that case (0).
 *
 * also note that seeking relative to the "end of file" isn't supported:
 * it has no meaning, so it returns -EINVAL.
 */
static int flash_lseek(struct inode * inode, struct file * file, off_t offset, int orig)
{

	if (flashdebug)
	printk("Flash_dev: flash_lseek, offset=0x%X, orig=0x%X.\n",
	    (unsigned int)offset, (unsigned int)orig);

	switch (orig) {
		case 0:
			if (offset > gbFlashSize)
			    return -EINVAL;
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			if ((file->f_pos + offset) > gbFlashSize)
			    return -EINVAL;
			file->f_pos += offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
	if (file->f_pos < 0)
	{
		return 0;
	}

	return file->f_pos;
}


//assume that main Write routine did the parameter checking...
//so just go ahead and erase, what requested!

#ifndef MINIKERNEL
static
#endif
int erase_block(int nBlock)
{
volatile unsigned int	c1;
volatile unsigned char* pWritePtr;
int temp,temp1;
unsigned long	flags;


        setLED(3,(int)&flash_miscdev);	//orange LED == erase
	
	// reset footbridge to the correct offset 0 (...0..3)
	*(volatile unsigned char*)KROM_ACCESS = 0;

	c1 = *(volatile unsigned char *)(KFLASH_BASE+0x8000);	//dummy ROM read
	kick_open();
	*(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0x50;	//reset status if old errors

	// erase a block...
	//aim at the middle of a current block...
	pWritePtr = (unsigned char*)((unsigned int)(KFLASH_BASE+0x8000+(nBlock<<16)));
	c1 = *pWritePtr;	//dummy read
	
	save_flags(flags);
	cli();
	kick_open();
	*(volatile unsigned char*)pWritePtr = 0x20;	//erase
	*(volatile unsigned char*)pWritePtr = 0xD0;	//confirm
	restore_flags(flags);

	//wait 10 ms
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + 1;	//wait 10 ms
	schedule();
			
	// wait while erasing in process (up to 10 sec)
	temp = jiffies+10*HZ;
	c1=0;
	current->timeout = jiffies + 1;	//wait 10 ms
	while (!(c1 & 0x80) && temp>jiffies)
	{
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + 1;	//wait 10 ms
		schedule();
		c1 = *(volatile unsigned char*)(pWritePtr);	//*(unsigned char *)(KFLASH_BASE+0x18000); //read any address
//printk("Flash_erase: status=%X.\n",c1);
	}

	//set flash for normal read access
	kick_open();
//	*(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0xFF;
	*(volatile unsigned char*)pWritePtr = 0xFF; //back to normal operation

	//check if erase errors were reported
	if (c1 & 0x20)
	{
		if (flashdebug)
		printk("Flash_erase: err at %X.\n",(unsigned int)pWritePtr);
		*(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0x50;	//reset error
		return(-2);
	}
			

	// just for testing - verify if erased OK...

	pWritePtr = (unsigned char*)((unsigned int)(KFLASH_BASE+(nBlock<<16)));
	for (temp=0;temp<16*1024;temp++,pWritePtr+=4)
	{
		if ((temp1 = *(volatile unsigned int*)pWritePtr) != 0xFFFFFFFF)
		{
			if (flashdebug)
			printk("Flash_erase: verify err at %X = %X.\n",
			    (unsigned int)pWritePtr,temp1);
			return(-1);
		}
	}

	return(0);		
		
}		

//write_block will limit number of bytes written to the space in this block

#ifndef MINIKERNEL
static
#endif
int write_block(unsigned long p, const char* buf, int count)
{
volatile unsigned int	c1,c2;
unsigned char* pWritePtr;
unsigned int uAddress;
unsigned int offset, timeout, timeout1;
unsigned long	flags;


        setLED(2,(int)&flash_miscdev);	//red LED == write
	
	pWritePtr = (unsigned char*)((unsigned int)(KFLASH_BASE + p));

	offset = p & 0xFFFF; //check if write will end in this block....
	if (offset + count > 0x10000)
	{
	    count = 0x10000-offset;
	}
	
	for (offset=0;offset<count;offset++,pWritePtr++)
	{
		uAddress = (unsigned int)pWritePtr;
		uAddress &= 0xFFFFFFFC;
		c2 = get_user_byte(buf+offset);
		timeout = jiffies+30*HZ;	//wait up to 30 sec for this byte

WriteRetry:		
		c1 = *(volatile unsigned char *)(KFLASH_BASE+0x8000);//dummy read
				
		save_flags(flags);
		cli();

		//kick open the write gate 
		kick_open();
		
		// program footbridge to the correct offset...0..3
		*(volatile unsigned char*)(KROM_ACCESS) = (unsigned int)pWritePtr & 3;
		
		*(volatile unsigned char*)(uAddress) = 0x40;	//write cmd
		*(volatile unsigned char*)(uAddress) = c2;//data to write
		
		*(volatile unsigned char*)(KFLASH_BASE+0x10000) = 0x70; //get status
		restore_flags(flags);

		c1=0;
		timeout1 = jiffies+1*HZ;	//wait up to 1 sec for this byte
		while (!(c1 & 0x80) && timeout1>jiffies)	//while not ready...
		{
			c1 = *(volatile unsigned char *)(KFLASH_BASE+0x8000);
		}
		
		if (timeout1<=jiffies)	//if timeout getting status
		{
		    kick_open();
		    *(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0x50; //reset err
		    goto WriteRetry;
		}

		//switch on read access, as a default flash operation mode
		kick_open();
		*(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0xFF; //read access
		
	//if hardware reports an error writing, and not timeout - 
	// reset the chip and retry
				
		if (c1 & 0x10)	//the byte write status BAD?
		{
		    kick_open();
		    *(volatile unsigned char*)(KFLASH_BASE+0x8000) = 0x50; //reset err
		    
		    if (timeout <= jiffies)	//before timeout?
		    {
//			if (flashdebug)
			{
			    printk("FlashWrite: Retrying write (addr=0x%X)...\n",
			        (unsigned int)pWritePtr-KFLASH_BASE);
			}	

		        setLED(0,(int)&flash_miscdev);	//no LED == waiting
			schedule();	//wait 10 ms		
		        setLED(2,(int)&flash_miscdev);	//red LED == write
			goto WriteRetry;
		    }
		    else
		    {	
			printk("Timeout in flash write! (addr=0x%X) Aborting...\n",
			    (unsigned int)pWritePtr-KFLASH_BASE);
			return(-2); //return error -2
		    }
		}
	}


        setLED(1,(int)&flash_miscdev);	//green LED == read/verify

	pWritePtr = (unsigned char*)((unsigned int)(KFLASH_BASE + p));
	for (offset=0;offset<count;offset++)
	{
	    if (*pWritePtr++ != get_user_byte(buf++))
	    {
		if (flashdebug)
		printk("flash write verify error at 0x%X! Retrying...\n",
		    (unsigned int)pWritePtr);
		return(0);
	    }
	}

	return(count);
}


#ifndef MINIKERNEL
static
#endif
void kick_open()
{
unsigned int timer15us;

// if we use Rev5 timer 1  (its is a shaken +-25% 66Mhz), 15us == ~1000 tiks
// if we go for an exact timer 3 (7 Mhz), tick=0.1428us, so we need 100+ ticks

//	if (rev < 4) return;	//on rev.3 no need to do this...
	
	//we want to write a bit pattern XXX1 to Xilinx to enable write
	//the write gate will be open for the next ~2ms.
	set_CPLD(0x01,0x01);

	//we need to compensate for the ISA delay - put a 15us wait
	timer15us = *(volatile unsigned int*)FB_TIMER1;
	timer15us -= TIMER1_15us;	//decrement by 1000 == 15us for 304
	while(timer15us < *(volatile unsigned int*)FB_TIMER1) {};
}

int nwflash_init(void)
{
int id;

    id=get_flash_id();
    printk("Flash ROM driver v.5.1, h/w flash ID 0x%04X.\n", id);

    misc_register(&flash_miscdev);
    return(0);
}

#ifdef MODULE
int init_module(void)
{
	return(nwflash_init());
}

void cleanup_module(void)
{
	misc_deregister(&flash_miscdev);
}
#endif
