/*
 * linux/arch/arm/lib/irqs-vnc.c
 *
 * Copyright 1998 Corel Computer Corp
 *
 * experiment to see if we can produce C code, not needing .S code
 */

#include <linux/config.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/ptrace.h>
#include <linux/interrupt.h>

extern int do_IRQ(int, struct pt_regs*);

int bad_IRQ_interrupt(int irq_no, struct pt_regs* regs)
{
    /* block it, and tell the user */
    if (irq_no>=16)
        ; /* these interrupts will not fire again until cleared */
    else
        /* mask off the interrupt in the PIC chip */
        if (irq_no<8)
            *(volatile unsigned char*)(IO_BASE+PIC_MASK_LO) |= 1 << irq_no;
        else
            *(volatile unsigned char*)(IO_BASE+PIC_MASK_HI) |= 1 << (irq_no-8);
            
    printk("Bad interrupt %d received!\n",irq_no);
    return 0;
}

int fast_IRQ_interrupt(int irq_no, struct pt_regs* regs)
{
    /* do the footbridge group */
    if (irq_no>=16) {
        volatile int* p = (volatile int*)(IO_BASE_ARM_CSR);
        /* disable the interrupt */
        p[IRQ_ENABLE_CLEAR/4] = 1 << (irq_no-16);
        intr_count++;
        do_IRQ(irq_no, regs);
        cli();
        /* re-enable this interrupt */
        p[IRQ_ENABLE_SET/4] = 1 << (irq_no-16);
    } else {
        volatile unsigned char* p = (volatile unsigned char*)IO_BASE;
        /* disable the interrupt */
        if (irq_no<8) {
            p[PIC_MASK_LO] |= 1 << (irq_no-0);
            p[PIC_LO] = 0x20; /* ack the interrupt */
        } else {
            p[PIC_MASK_HI] |= 1 << (irq_no-8);
            p[PIC_HI] = 0x20; /* ack the interrupt */
            p[PIC_LO] = 0x20; /* and the chained device */
        }
        intr_count++;
        do_IRQ(irq_no, regs);
        cli();

        /* re-enable the interrupt */
        if (irq_no<8)
            p[PIC_MASK_LO] &= ~(1 << (irq_no));
        else
            p[PIC_MASK_HI] &= ~(1 << (irq_no-8));
    }
    intr_count--;
    return 1;
}

int IRQ_interrupt(int irq_no, struct pt_regs* regs)
{
    /* do the footbridge group */
    if (irq_no>=16) {
        volatile int* p = (volatile int*)(IO_BASE_ARM_CSR);
        /* disable the interrupt */
        p[IRQ_ENABLE_CLEAR/4] = 1 << (irq_no-16);
        intr_count++;
        sti();
        do_IRQ(irq_no, regs);
        cli();
        /* re-enable this interrupt */
        p[IRQ_ENABLE_SET/4] = 1 << (irq_no-16);
    } else {
        volatile unsigned char* p = (volatile unsigned char*)IO_BASE;
        /* disable the interrupt */
        if (irq_no<8) {
            p[PIC_MASK_LO] |= 1 << (irq_no);
            p[PIC_LO] = 0x20; /* ack the interrupt */
        } else {
            p[PIC_MASK_HI] |= 1 << (irq_no-8);
            p[PIC_HI] = 0x20; /* ack the interrupt */
            p[PIC_LO] = 0x20; /* and the chained device */
        }
        intr_count++;
        sti();
        do_IRQ(irq_no, regs);
        cli();

        /* re-enable the interrupt */
        if (irq_no<8)
            p[PIC_MASK_LO] &= ~(1 << (irq_no));
        else
            p[PIC_MASK_HI] &= ~(1 << (irq_no-8));
    }
    intr_count--;
    return 0;
}

timer_IRQ_interrupt(int irq_no, struct pt_regs* regs)
{
    int flags;
    /* do the footbridge group; TIMER1 or TIMER3 */
    if (irq_no>=16) {
        volatile int* p = (volatile int*)(IO_BASE_ARM_CSR);
        /* disable the interrupt */
        p[IRQ_ENABLE_CLEAR/4] = 1 << (irq_no-16);
        intr_count++;
        save_flags(flags);
        do_IRQ(irq_no, regs);
        restore_flags(flags);
        /* re-enable this interrupt */
        p[IRQ_ENABLE_SET/4] = 1 << (irq_no-16);
    } else {
        volatile unsigned char* p = (volatile unsigned char*)IO_BASE;
        /* disable the interrupt */
        if (irq_no<8) {
            p[PIC_MASK_LO] |= 1 << (irq_no);
            p[PIC_LO] = 0x20; /* ack the interrupt */
        } else {
            p[PIC_MASK_HI] |= 1 << (irq_no-8);
            p[PIC_HI] = 0x20; /* ack the interrupt */
            p[PIC_LO] = 0x20; /* and the chained device */
        }
        intr_count++;
        save_flags(flags);
        do_IRQ(irq_no, regs);
        restore_flags(flags);

        /* re-enable the interrupt */
        if (irq_no<8)
            p[PIC_MASK_LO] &= ~(1 << (irq_no));
        else
            p[PIC_MASK_HI] &= ~(1 << (irq_no-8));
    }
    intr_count--;
    return 0;
}

int probe_IRQ_interrupt(int irq_no, struct pt_regs* regs)
{
    /* do the footbridge group */
    if (irq_no>=16){
        volatile int* p = (volatile int*)(IO_BASE_ARM_CSR);
        /* disable the interrupt */
        p[IRQ_ENABLE_CLEAR/4] = 1 << (irq_no-16);
    } else {
        volatile unsigned char* p = (volatile unsigned char*)IO_BASE;
        /* disable the interrupt */
        if (irq_no<8) {
            p[PIC_MASK_LO] |= 1 << (irq_no);
            p[PIC_LO] = 0x20; /* ack the interrupt */
        } else {
            p[PIC_MASK_HI] |= 1 << (irq_no-8);
            p[PIC_HI] = 0x20; /* ack the interrupt */
            p[PIC_LO] = 0x20; /* and the chained device */
        }
    }
    return 0;
}
