#ifndef __ARM_UNALIGNED_H
#define __ARM_UNALIGNED_H

#include <asm/segment.h>

#define get_unaligned(ptr) \
	((__typeof__(*(ptr)))__get_unaligned((ptr), sizeof(*(ptr))))

#define put_unaligned(val, ptr) \
	__put_unaligned((unsigned long)(val), (ptr), sizeof(*(ptr)))

extern void bad_unaligned_access_length (void);
	
extern inline unsigned long __get_unaligned(const void *ptr, size_t size)
{
    unsigned long val;
    switch (size) {
    case 1:
        val = get_user_byte(ptr);
        break;

    case 2:
        val = get_user_byte(ptr)
            | (get_user_byte(ptr+1)<<8);
        break;

    case 4:
        val = get_user_byte(ptr)
            | (get_user_byte(ptr+1)<<8)
            | (get_user_byte(ptr+2)<<16)
            | (get_user_byte(ptr+3)<<24);
	    break;

	default:
	    bad_unaligned_access_length ();
    }
    return val;
}

extern inline void __put_unaligned(unsigned long val, void *ptr, size_t size)
{
    switch (size) {
    case 1:
        put_user_byte(val, ptr);
        break;

    case 2:
        put_user_byte(val, ptr);
        put_user_byte(val>>8, ptr+1);
        break;

    case 4:
        put_user_byte(val, ptr);
        put_user_byte(val>>8, ptr+1);
        put_user_byte(val>>16, ptr+2);
        put_user_byte(val>>24, ptr+3);
        break;

    default:
        bad_unaligned_access_length ();
    }
}

#endif
